%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% 20bit Mem, blind, T = 200, K = 8000 %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% H. Jaeger, Feb 02, 2012
%
% This script runs Ntrials many complete training-testing experiments, each
% using fresh data and new random reservoirs.

addpath('../share');
set(0,'DefaultFigureWindowStyle','normal');

saveFlag = 1; plotFlag = 0;

newModel = 1; newData = 1;

Ntrials = 100; % how many complete train-test trials are to be done.

%%% Set global parameters for a single train-test experiment
T = 200; %  length of trial sequences
Ntrain = 500; Ntest = 1000;
D = 5; % nr of payload symbols
M = 10; % length of memory sequence
nInternalUnits = 8000; % Size of reservoir
connectivity = 0.001333; % average connectivity of reservoir
% (1 = fully connected)
alpha = 1e-20;
NpickTrain = 30; % how many states from the interim period are used for training
inScaling = 0.0000001*[1 1 1 1 1 1 1]'; % how the two sets of input weights are scaled
SR = 1; % Weight matrix spectral radius
LR = 1; % smoothing rate, from (0,1]. Low values = much smoothing.
% Value of 1: no smoothing, i.e. reservoir becomes standard
% sigmoid unit network

%%% Create diagnostic collector arrays
genTimeCollector = zeros(1,Ntrials);
learnTimeCollector = zeros(1,Ntrials);
trainAvAbsErrCollector = zeros(1,Ntrials);
WoutAvAbsSizeCollector = zeros(1,Ntrials);
maxAbsErrTrainCollector = zeros(1,Ntrials);
misClassNrCollector = zeros(1,Ntrials);
trialFailCollector = zeros(1,Ntrials);
runFailCollector = zeros(1,Ntrials);
testAvAbsErrCollector = zeros(1,Ntrials);
maxAbsErrTestCollector = zeros(1,Ntrials);



%%% run experiments
for trialNr = 1:Ntrials
    tAllVar = tic;
    fprintf('%g ', trialNr);
    if mod(trialNr, 20) == 0
        fprintf('\n');
    end
    
    % Generate training and testing data
    if newData
        datasetTrain = genDataMem(T, D, M, Ntrain);
        datasetTest = genDataMem(T, D, M, Ntest);
    end
    
    % Generate reservoir
    if newModel
        tGenVar = tic;
        WunitSR = ...
            generate_internal_weights(nInternalUnits, connectivity);
        tGen = toc(tGenVar);
        WinRaw = 2 * (rand(nInternalUnits, 7) - 0.5);
    end
    
    % Scale weighgts
    W = SR * WunitSR;
    Win = WinRaw * diag(inScaling);
    
    startstate = zeros( nInternalUnits, 1);
    
    % start stopwatch for training
    tTrainVar = tic;
    
    NP = 22+NpickTrain;
    R = zeros(nInternalUnits, nInternalUnits);
    ST = zeros(nInternalUnits, 7);
    
    for i = 1:Ntrain
        pickStart = mod(NpickTrain*i,T-2);
        if pickStart == 0
            pickStart = T-2;
        end
        pickEnd = pickStart + NpickTrain - 1;
        if pickEnd <= T-2
            pickInterval = (pickStart:1:pickEnd) + 11;
        else
            pickInterval = [1:pickEnd-(T-2), pickStart:1:(T-2)] + 11;
        end
        pickindices = [1:11, pickInterval, T+10:T+20];
        thisInput = datasetTrain(:, 1:7, i);
        stateHarvestTrain = ...
            pickstates(W, Win, LR, thisInput, startstate, pickindices);
        teachHarvestTrain = ...
            datasetTrain(pickindices,8:14,i)';
        R = R + stateHarvestTrain * stateHarvestTrain';
        ST = ST + stateHarvestTrain * teachHarvestTrain';
    end
    
    
    % Compute output weights
    Wout = (inv(R + alpha * eye(nInternalUnits)) * ST)';
    
    tTrain = toc(tTrainVar);
    
    % Testing
    NP = T+2*M;
    teachHarvestTest = zeros(D+2, NP*Ntest);
    testOuts = zeros(D+2, NP*Ntest);
    failCounter = 0;
    for i = 1:Ntest
        thisInput = datasetTest(:, 1:(D+2), i);
        stateHarvest = ...
            allstatesNotstartstateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvestTest(:,NP*(i-1)+1:NP*i) = datasetTest(:,(D+3):end,i)';
        thisTestOut = Wout * stateHarvest;
        testOuts(:,NP*(i-1)+1:NP*i) = thisTestOut;
        % count as fail if not perfect
        if sum(sum(abs(thisTestOut - datasetTest(:,(D+3):end,i)') > 0.5))...
                > 0
            failCounter = failCounter + 1;
        end
    end
    
    % Performance diagnostics for this trial
    absErrorsTest = abs(teachHarvestTest - testOuts);
    avAbsErrTest = mean(mean(absErrorsTest));
    maxAbsErrTest = max(max(absErrorsTest));
    nrMisclassifieds = sum(sum(absErrorsTest >= 0.5));
    
    % Save diagnostics
    genTimeCollector(trialNr) = tGen;
    learnTimeCollector(trialNr) = tTrain;
    runFailCollector(trialNr) = failCounter;
    testAvAbsErrCollector(trialNr) = avAbsErrTest;
    WoutAvAbsSizeCollector(trialNr) = mean(mean(abs(Wout)));
    maxAbsErrTestCollector(trialNr) = maxAbsErrTest;
    misClassNrCollector(trialNr) = nrMisclassifieds;
    if maxAbsErrTest > 0.5
        trialFailCollector(trialNr) = 1;
    end
    
    
    %%% save results
    if saveFlag
        save('mem20bitResultsT200Units8000Blind', 'learnTimeCollector',...
            'genTimeCollector',  ...
            'testAvAbsErrCollector', 'maxAbsErrTestCollector', ...
            'WoutAvAbsSizeCollector',  ...
            'misClassNrCollector', 'trialFailCollector', 'trialNr');
    end
    
    tAll = toc(tAllVar); % tAll is about 20 minutes here
end


%%% display results
fprintf('\nmean generation time = %g', mean(genTimeCollector));
fprintf('\nmean learning time = %g', mean(learnTimeCollector));
fprintf('\nmean av abs error test = %g', mean(testAvAbsErrCollector));
fprintf('\nmean abs Wout size = %g', mean(WoutAvAbsSizeCollector));
fprintf('\nmax abs err test = %g', max(maxAbsErrTestCollector));
fprintf('\nMax misclassifieds per trial = %g', max(misClassNrCollector));
fprintf('\nNr failed trials = %g\n\n', sum(misClassNrCollector > 0));


if plotFlag
    
    % plot maxErrors
    figure(1); clf;
    set(gcf, 'Position', [900 700 300 250]);
    plot(sort(log2(maxAbsErrTestCollector)), 'LineWidth',2);
    %title('log2 max abs errors test, sorted');
    set(gca, 'FontSize', 12);
    xlabel('trials', 'FontSize', 12);
    ylabel('log2 max abs error', 'FontSize', 12);
    
    % plot dependency of error on WoutSize
    figure(2); clf;
    set(gcf, 'Position', [900 450 300 250]);
    plot(log2(WoutAvAbsSizeCollector), log2(maxAbsErrTestCollector),'x', ...
        'LineWidth',1.5);
    %title('log2 maxAbsErr vs. WoutSize');
    set(gca, 'FontSize', 12);
    xlabel('log2 av abs Wout', 'FontSize', 12);
    ylabel('log2 max abs error', 'FontSize', 12);
    
    
    % plot exemplary state sequences. If you want to display more than one,
    % repeat execution of the cell below.
    runNr = 0; figNr = 2;
    %%
    runNr = runNr+1;
    
    thisInput = datasetTrainTest(:, 1:7, runNr);
    stateHarvestPlot = ...
        allstatesNotstartstateLeaky(W, Win, LR, thisInput, startstate);
    figNr = figNr + 1;
    figure(figNr);
    set(gcf, 'Position', [1250 700 400 250]);
    subplot(1,2,1);
    plot(1:205, stateHarvestPlot(1:8,1:205)', 'LineWidth',2);
    ylabel('unit activation', 'FontSize', 12);
    % title(sprintf(['Example state sequence']));
    subplot(1,2,2);
    plot(206:220, stateHarvestPlot(1:8,206:end)', 'LineWidth',2);
    xlabel('time (network updates)', 'FontSize', 12);
end


