%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The 5bit Mem task, basic, T = 200, K = 500 %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% H. Jaeger, Jan 31, 2012
%
% This script runs Ntrials many complete training-testing experiments, each
% using fresh data and new random reservoirs.

addpath('../share');
set(0,'DefaultFigureWindowStyle','normal');

saveFlag = 1;

newModel = 1; newData = 0;

Ntrials = 100; % how many complete train-test trials are to be done.

%%% Set global parameters for a single train-test experiment
T = 200; %  length of trial sequences
nInternalUnits = 500; % Size of reservoir
connectivity = 0.02; % average connectivity of reservoir
% (1 = fully connected)
inScaling = 0.0002*[1 1 .01 1000]'; % how the two sets of input weights are scaled
SR = 1; % Weight matrix spectral radius
LR = 1; % smoothing rate, from (0,1]. Low values = much smoothing.
% Value of 1: no smoothing, i.e. reservoir becomes standard
% sigmoid unit network

%%% Create diagnostic collector arrays
genTimeCollector = zeros(1,Ntrials);
learnTimeCollector = zeros(1,Ntrials);
trainAvAbsErrCollector = zeros(1,Ntrials);
WoutAvAbsSizeCollector = zeros(1,Ntrials);
maxAbsErrTrainCollector = zeros(1,Ntrials);
misClassNrCollector = zeros(1,Ntrials);
trialFailCollector = zeros(1,Ntrials);

%%% run experiments
for trialNr = 1:Ntrials
    tAllVar = tic;
    fprintf('%g ', trialNr);
    if mod(trialNr, 20) == 0
        fprintf('\n');
    end
    
    % Generate training and testing data
    if newData
        datasetTrainTest = genDataMem_All(T, 2, 5);
        Ntrain = size(datasetTrainTest,3);
        % in this small-sample deterministic task, train data  =  test data
    end
    
    % Generate reservoir
    if newModel
        tGenVar = tic;
        WunitSR = generate_internal_weights(nInternalUnits, connectivity);
        tGen = toc(tGenVar);
        WinRaw = 2 * (rand(nInternalUnits, 4) - 0.5);
    end
    
    % Scale weighgts
    W = SR * WunitSR;
    Win = WinRaw * diag(inScaling);
    
    startstate = zeros( nInternalUnits, 1);
    
    % start stopwatch for training
    tTrainVar = tic;
    
    % Collect training state vectors.
    NP = T+10; % length of a run
    stateHarvest = zeros(nInternalUnits, NP*Ntrain);
    teachHarvest = zeros(4, NP*Ntrain);
    for i = 1:Ntrain
        thisInput = datasetTrainTest(:, 1:4, i);
        stateHarvest(:,NP*(i-1)+1:NP*i) = ...
            allstatesNotstartstateLeaky(W, Win, LR, ...
            thisInput, startstate);
        teachHarvest(:,NP*(i-1)+1:NP*i) = ...
            datasetTrainTest(:,5:8,i)';
    end
    
    % Compute output weights
    Wout = (pinv(stateHarvest') * teachHarvest')';
    
    tTrain = toc(tTrainVar);
    
    % Error diagnostics
    trainOuts = Wout * stateHarvest;
    trainAbsErrs = abs(trainOuts - teachHarvest);
    nrMisclassifieds = sum(sum(trainAbsErrs > 0.5));
    
    
    % Save diagnostics
    genTimeCollector(trialNr) = tGen;
    learnTimeCollector(trialNr) = tTrain;
    trainAvAbsErrCollector(trialNr) = mean(mean(trainAbsErrs));
    maxAbsErrTrainCollector(trialNr) = max(max(trainAbsErrs));
    WoutAvAbsSizeCollector(trialNr) = mean(mean(abs(Wout)));
    misClassNrCollector(trialNr) = nrMisclassifieds;
    if nrMisclassifieds > 0 % notice that here trial is failed if
        % not all runs are perfect
        trialFailCollector(trialNr) = 1;
    end
    
    
    tAll = toc(tAllVar); % tAll is about 4 secs here
end

%%% save results
if saveFlag
    save('mem5bitResultsT200Units500Basic', 'learnTimeCollector',...
        'genTimeCollector', 'trainAvAbsErrCollector', ...
        'WoutAvAbsSizeCollector', 'maxAbsErrTrainCollector', ...
        'misClassNrCollector', 'trialFailCollector');
end

%%% display results
fprintf('\nmean learning time = %g', mean(learnTimeCollector));
fprintf('\nmean av abs error train = %g', mean(trainAvAbsErrCollector));
fprintf('\nmean abs Wout size = %g', mean(WoutAvAbsSizeCollector));
fprintf('\nmax abs err train = %g', max(maxAbsErrTrainCollector));
fprintf('\nNr misclassifieds = %g', sum(misClassNrCollector));
fprintf('\nMax misclassifieds per trial = %g', max(misClassNrCollector));
fprintf('\nNr failed trials = %g\n\n', sum(trialFailCollector));

% plot maxErrors
figure(1); clf;
set(gcf, 'Position', [900 700 300 250]);
plot(sort(log2(maxAbsErrTrainCollector)), 'LineWidth',2);
%title('log2 max abs errors test, sorted');
set(gca, 'FontSize', 12);
xlabel('trials', 'FontSize', 12);
ylabel('log2 max abs error', 'FontSize', 12);

% plot dependency of error on WoutSize
figure(2); clf; 
set(gcf, 'Position', [900 700 300 250]);
plot(log2(WoutAvAbsSizeCollector), log2(maxAbsErrTrainCollector),'x', ...
    'LineWidth',1.5);
%title('log2 maxAbsErr vs. WoutSize');
set(gca, 'FontSize', 12);
xlabel('log2 av abs Wout', 'FontSize', 12);
ylabel('log2 max abs error', 'FontSize', 12);


% plot exemplary state sequences. If you want to display more than one,
% repeat execution of the cell below.
runNr = 0; figNr = 2;
%%
runNr = runNr+1;

thisInput = datasetTrainTest(:, 1:4, runNr);
stateHarvestPlot = ...
    allstatesNotstartstateLeaky(W, Win, LR, thisInput, startstate);
figNr = figNr + 1;
figure(figNr);
set(gcf, 'Position', [1250 700 400 250]);
subplot(1,2,1);
plot(stateHarvestPlot(1:8,1:200)', 'LineWidth',2);
ylabel('unit activation', 'FontSize', 12);
% title(sprintf(['Example state sequence']));
subplot(1,2,2);
plot(201:210, stateHarvestPlot(1:8,201:end)', 'LineWidth',2);
xlabel('time (network updates)', 'FontSize', 12);


