%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% 5bit Mem task, smart, T = 1000, K = 500 %%%%%%
%%% Version used in Discussion section %%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% H. Jaeger, June 19, 2014
%
% This script runs Ntrials many complete training-testing experiments, each
% using fresh data and new random reservoirs.

addpath('../share');
set(0,'DefaultFigureWindowStyle','normal');

randstate = 1;
% Initializations for random numbers
randn('state', randstate);
rand('twister', randstate);


saveFlag = 0;

newModel = 1; newData = 1;

Ntrials = 1; % how many complete train-test trials are to be done.

%%% Set global parameters for a single train-test experiment
T = 1000; %  length of trial sequences
nInternalUnits = 220; % Size of reservoir
connectivity = 0.02; % average connectivity of reservoir
% (1 = fully connected)
periodRange = [2 10];
oscNr = 100;
inScaling = 0.0002*[1 1 .0 1000]'; % how the two sets of input weights are scaled
SR = 1; % Weight matrix spectral radius
LR = 1; % smoothing rate, from (0,1]. Low values = much smoothing.
% Value of 1: no smoothing, i.e. reservoir becomes standard
% sigmoid unit network
WoutType = 4; % 1: learn from entire reservoir
% 2: learn from oscillators only
% 3: learn from non-oscillators only
% 4: learn from oscillators only, plus a static nonlinear
%    transform of them
equalizeAmplitudes = 1; %whether oscillator amplitudes are made equal after cue

%%% Create diagnostic collector arrays
genTimeCollector = zeros(1,Ntrials);
learnTimeCollector = zeros(1,Ntrials);
trainAvAbsErrCollector = zeros(1,Ntrials);
WoutAvAbsSizeCollector = zeros(1,Ntrials);
maxAbsErrTrainCollector = zeros(1,Ntrials);
misClassNrCollector = zeros(1,Ntrials);
trialFailCollector = zeros(1,Ntrials);

%%% run experiments
for trialNr = 1:Ntrials
    tAllVar = tic;
    fprintf('%g ', trialNr);
    if mod(trialNr, 20) == 0
        fprintf('\n');
    end
    
    % Generate training and testing data
    if newData
        datasetTrainTest = genDataMem_All(T, 2, 5);
        Ntrain = size(datasetTrainTest,3);
        % in this small-sample deterministic task, train data  =  test data
    end
    
    % Generate reservoir
    if newModel
        tGenVar = tic;
        WunitSR = ...
            genOscillatorReservoir(periodRange, ...
            oscNr, nInternalUnits, connectivity);
        tGen = toc(tGenVar);
        WinRaw = 2 * (rand(nInternalUnits, 4) - 0.5);
    end
    
    % Scale weighgts
    W = SR * WunitSR;
    Win = WinRaw * diag(inScaling);
    
    startstate = zeros( nInternalUnits, 1);
    
    % start stopwatch for training
    tTrainVar = tic;
    
    % Collect training state vectors.
    NP = T+10; % length of a run
    stateHarvest = zeros(nInternalUnits, NP*Ntrain);
    teachHarvest = zeros(4, NP*Ntrain);
    stateHarvest1 = zeros(nInternalUnits, NP,Ntrain);
    teachHarvest1 = zeros(4, NP,Ntrain);
    unitInds = 1:2*oscNr;
    EqTime = 500;
    if equalizeAmplitudes
        for i = 1:Ntrain
            thisInput = datasetTrainTest(:, 1:4, i);
            thisHarvest = ...
                allstatesNotstartstateLeakyEq(W, Win, LR, ...
                thisInput, startstate, unitInds, EqTime);
            stateHarvest(:,NP*(i-1)+1:NP*i) = thisHarvest;
            teachHarvest(:,NP*(i-1)+1:NP*i) = ...
                datasetTrainTest(:,5:8,i)';
            
            stateHarvest1(:,:,i) = thisHarvest;
            teachHarvest1(:,:,i) = ...
                datasetTrainTest(:,5:8,i)';
        end
    else
        for i = 1:Ntrain
            thisInput = datasetTrainTest(:, 1:4, i);
            thisHarvest = ...
                allstatesNotstartstateLeaky(W, Win, LR, ...
                thisInput, startstate);
            stateHarvest(:,NP*(i-1)+1:NP*i) = thisHarvest;
            teachHarvest(:,NP*(i-1)+1:NP*i) = ...
                datasetTrainTest(:,5:8,i)';
            
            stateHarvest1(:,:,i) = thisHarvest;
            teachHarvest1(:,:,i) = ...
                datasetTrainTest(:,5:8,i)';
        end
    end
    
    % Compute output weights
    if WoutType == 1
        Wout = (pinv(stateHarvest') * teachHarvest')';
    elseif WoutType == 2
        Wout = (pinv(stateHarvest(1:2*oscNr,:)') * teachHarvest')';
    elseif WoutType == 3
        Wout = (pinv(stateHarvest( 2*oscNr + 1:end,:)') * teachHarvest')';
    elseif WoutType == 4
        Wout = (pinv([stateHarvest(1:2*oscNr,:);...
            tanh(stateHarvest(1:2*oscNr,:))]') * teachHarvest')';
    end
    %       Wout = ...
    %         (inv(stateHarvest * stateHarvest' + 0.0000000001 * eye(nInternalUnits)) ...
    %         * stateHarvest * teachHarvest')';
    
    tTrain = toc(tTrainVar);
    
    % Error diagnostics according to original task
    if WoutType == 1
        trainOuts = Wout * stateHarvest;
    elseif WoutType == 2
        trainOuts = Wout * stateHarvest(1:2*oscNr,:);
    elseif WoutType == 3
        trainOuts = Wout * stateHarvest( 2*oscNr + 1:end,:);
    elseif WoutType == 4
        trainOuts = Wout * [stateHarvest(1: 2*oscNr,:);...
            tanh(stateHarvest(1: 2*oscNr ,:))];
    end
    trainAbsErrs = abs(trainOuts - teachHarvest);
    nrMisclassifieds = sum(sum(trainAbsErrs > 0.5));
    
    % Error diagnostics based only on 5-bit pattern recall part
    trainAbsErrs1 = zeros(2,5,Ntrain);
    nrMisclassifieds1 = 0;
    summedAbsErrs1 = 0;
    maxAbsErrs1 = 0;
    for i = 1:Ntrain
        if WoutType == 1
            theseOuts = Wout * stateHarvest1(:,end-4:end,i);
        elseif WoutType == 2
            theseOuts = Wout * stateHarvest1(1:2*oscNr,end-4:end,i);
        elseif WoutType == 3
            theseOuts = Wout * stateHarvest1( 2*oscNr + 1:end,end-4:end,i);
        elseif WoutType == 4
            theseOuts = Wout * [stateHarvest1(1:2*oscNr,end-4:end,i);...
                tanh(stateHarvest1(1:2*oscNr,end-4:end,i))];
        end
        theseTeachs = teachHarvest1(:,end-4:end,i);
        trainAbsErrs1(:,:,i) = abs(theseOuts(1:2,:) - theseTeachs(1:2,:));
        nrMisclassifieds1 = nrMisclassifieds1 + ...
            sum(sum(trainAbsErrs1(:,:,1) > 0.5));
        summedAbsErrs1 = summedAbsErrs1 + ...
            sum(sum(trainAbsErrs1(:,:,1)));
        maxAbsErrs1 = max(maxAbsErrs1, max(max(trainAbsErrs1(:,:,1))));
    end
    
    
    
    % Save diagnostics
    genTimeCollector(trialNr) = tGen;
    learnTimeCollector(trialNr) = tTrain;
    trainAvAbsErrCollector(trialNr) = mean(mean(trainAbsErrs));
    maxAbsErrTrainCollector(trialNr) = max(max(trainAbsErrs));
    WoutAvAbsSizeCollector(trialNr) = mean(mean(abs(Wout)));
    misClassNrCollector(trialNr) = nrMisclassifieds;
    if nrMisclassifieds > 0 % notice that here trial is failed if
        % not all runs are perfect
        trialFailCollector(trialNr) = 1;
    end
    
    
    tAll = toc(tAllVar); % tAll is about 24 secs here
end

%%% save results
if saveFlag
    save('mem5bitResultsT1000Units500Smart', 'learnTimeCollector',...
        'genTimeCollector', 'trainAvAbsErrCollector', ...
        'WoutAvAbsSizeCollector', 'maxAbsErrTrainCollector', ...
        'misClassNrCollector', 'trialFailCollector');
end

%%% display results
fprintf('\nmean learning time = %g', mean(learnTimeCollector));
fprintf('\nmean av abs error train = %g', mean(trainAvAbsErrCollector));
fprintf('\nmean abs Wout size = %g', mean(WoutAvAbsSizeCollector));
fprintf('\nmax abs err train = %g', max(maxAbsErrTrainCollector));
fprintf('\nNr misclassifieds = %g', sum(misClassNrCollector));
fprintf('\nMax misclassifieds per trial = %g', max(misClassNrCollector));
fprintf('\nNr failed trials = %g\n', sum(trialFailCollector));
fprintf('********* only pattern recall ****');
fprintf('\nnrMisclassifieds1 = %g', nrMisclassifieds1);
fprintf('\nmeanAbsErrs1 = %g', summedAbsErrs1 / (320));
fprintf('\nmaxAbsErrs1 = %g\n\n', maxAbsErrs1);

% plot maxErrors
figure(1); clf;
set(gcf, 'Position', [900 700 300 250]);
plot(sort(log2(maxAbsErrTrainCollector)), 'LineWidth',2);
%title('log2 max abs errors test, sorted');
set(gca, 'FontSize', 12);
xlabel('trials', 'FontSize', 12);
ylabel('log2 max abs error', 'FontSize', 12);

% plot dependency of error on WoutSize
figure(2); clf;
set(gcf, 'Position', [900 450 300 250]);
plot(log2(WoutAvAbsSizeCollector), log2(maxAbsErrTrainCollector),'x', ...
    'LineWidth',1.5);
%title('log2 maxAbsErr vs. WoutSize');
set(gca, 'FontSize', 12);
xlabel('log2 av abs Wout', 'FontSize', 12);
ylabel('log2 max abs error', 'FontSize', 12);


% plot exemplary state sequences. If you want to display more than one,
% repeat execution of the cell below.

%%
runNr = 2;

thisInput = datasetTrainTest(:, 1:4, runNr);
if equalizeAmplitudes
    stateHarvestPlot = ...
    allstatesNotstartstateLeakyEq(W, Win, LR, thisInput, startstate, ...
     unitInds, EqTime);
else
stateHarvestPlot = ...
    allstatesNotstartstateLeaky(W, Win, LR, thisInput, startstate);
end
figure(3);
set(gcf, 'Position', [1250 700 400 250]);
subplot(1,2,1);
plot(1:1000, stateHarvestPlot(2*oscNr+1:2*oscNr+8,1:1000)', 'LineWidth',2);
ylabel('unit activation', 'FontSize', 12);
% title(sprintf(['Example state sequence']));
subplot(1,2,2);
plot(201:210, stateHarvestPlot(2*oscNr+1:2*oscNr+8,1001:end)', 'LineWidth',2);
xlabel('time (network updates)', 'FontSize', 12);

%%
figure(4);
set(gcf, 'Position', [1250 300 400 250]);
subplot(1,2,1);
plot(1:1000, stateHarvestPlot([1 3 5 7 9],1:1000)', 'LineWidth',2);
ylabel('unit activation', 'FontSize', 12);
% title(sprintf(['Example state sequence']));
subplot(1,2,2);
plot(201:210, stateHarvestPlot([1 3 5 7 9],1001:end)', 'LineWidth',2);
xlabel('time (network updates)', 'FontSize', 12);


