%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The multiplication task, blind, T = 500 %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% H. Jaeger, Jan 30, 2012
%
% This script runs Ntrials many complete training-testing experiments, each
% using fresh data and new random reservoirs.

addpath('../share');
set(0,'DefaultFigureWindowStyle','normal');

saveFlag = 1; plotFlag = 1;

newModel = 1; newData = 1;

Ntrials = 100; % how many complete train-test trials are to be done.

%%% Set global parameters for a single train-test experiment
T = 500; %  length of trial sequences
Ntrain = 2000;  % Nr of training sequences
Ntest = 1; % Nr of testing sequences
nInternalUnits = 1000; % Size of reservoir
connectivity = 0.01; % average connectivity of reservoir
% (1 = fully connected)
inScaling = 0.1*[1 1]'; % how the two sets of input weights are scaled
SR = 12; % Weight matrix spectral radius
LR = 0.0003; % smoothing rate, from (0,1]. Low values = much smoothing.
% Value of 1: no smoothing, i.e. reservoir becomes standard
% sigmoid unit network

%%% Create diagnostic collector arrays
genTimeCollector = zeros(1,Ntrials);
learnTimeCollector = zeros(1,Ntrials);
testAvAbsErrCollector = zeros(1,Ntrials);
trainAvAbsErrCollector = zeros(1,Ntrials);
WoutAvAbsSizeCollector = zeros(1,Ntrials);
maxAbsErrTestCollector = zeros(1,Ntrials);
maxAbsErrTrainCollector = zeros(1,Ntrials);
misClassNrCollector = zeros(1,Ntrials);
trialFailCollector = zeros(1,Ntrials);

%%% run experiments
for trialNr = 1:Ntrials
    tAllVar = tic;
    fprintf('%g ', trialNr);
    if mod(trialNr, 20) == 0
        fprintf('\n');
    end
    
    % Generate training and testing data
    if newData
        datasetTrain = genDataMult(T, Ntrain);
        datasetTest = genDataMult(T, Ntest);
    end
    
    % Generate reservoir
    if newModel
        tGenVar = tic;
        WunitSR = generate_internal_weights(nInternalUnits, connectivity);
        tGen = toc(tGenVar);
        WinRaw = 2 * (rand(nInternalUnits, 2) - 0.5);
    end
    
    % Scale weighgts
    W = SR * WunitSR;
    Win = WinRaw * diag(inScaling);
    
    startstate = zeros( nInternalUnits, 1);
    
    % start stopwatch for training
    tTrainVar = tic;
    
    % Collect training state vectors
    stateHarvest = zeros(nInternalUnits, Ntrain);
    teachHarvest = zeros(1, Ntrain);
    for i = 1:Ntrain
        thisRunlength = datasetTrain(1,3,i);
        thisInput = datasetTrain(1:thisRunlength, 1:2, i);
        stateHarvest(:,i) = ...
            laststateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvest(1,i) = datasetTrain(2,3,i);
    end
    
    % Compute output weights
    Wout = (pinv(stateHarvest') * teachHarvest')';
    
    tTrain = toc(tTrainVar);
    
    % Training error diagnostics
    trainOuts = Wout * stateHarvest;
    trainAbsErrs = abs(trainOuts - teachHarvest);
    
    % Compute test outputs
    stateHarvestTest = zeros(nInternalUnits, Ntest);
    teachHarvestTest = zeros(1, Ntest);
    
    for i = 1:Ntest
        thisRunlength = datasetTest(1,3,i);
        thisInput = datasetTest(1:thisRunlength, 1:2, i);
        stateHarvestTest(:,i) = laststateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvestTest(1,i) = datasetTest(2,3,i);
    end
    
    % Performance diagnostics for this trial
    testOuts = Wout * stateHarvestTest;
    avAbsErrTest = mean(abs(teachHarvestTest - testOuts));
    maxAbsErrTest = max(abs(teachHarvestTest - testOuts));
    nrMisclassifieds = sum(abs(teachHarvestTest - testOuts) > 0.04);
    
    
    % Save diagnostics
    genTimeCollector(trialNr) = tGen;
    learnTimeCollector(trialNr) = tTrain;
    trainAvAbsErrCollector(trialNr) = mean(trainAbsErrs);
    maxAbsErrTrainCollector(trialNr) = max(trainAbsErrs);
    testAvAbsErrCollector(trialNr) = avAbsErrTest;
    WoutAvAbsSizeCollector(trialNr) = mean(mean(abs(Wout)));
    maxAbsErrTestCollector(trialNr) = maxAbsErrTest;
    misClassNrCollector(trialNr) = nrMisclassifieds;
    if nrMisclassifieds > Ntest / 100
        trialFailCollector(trialNr) = 1;
    end
    
    
    tAll = toc(tAllVar); % tAll is little below 5 minutes here
end

%%% save results
if saveFlag
    save('multresultsT500Units1000Blind', 'learnTimeCollector',...
        'genTimeCollector', 'trainAvAbsErrCollector', ...
        'testAvAbsErrCollector', 'WoutAvAbsSizeCollector', ...
        'maxAbsErrTrainCollector', 'maxAbsErrTestCollector',...
        'misClassNrCollector', 'trialFailCollector');
end

%%% display results
fprintf('\nmean learning time = %g', mean(learnTimeCollector));
fprintf('\nmean av abs error train = %g', mean(trainAvAbsErrCollector));
fprintf('\nmean av abs error test = %g', mean(testAvAbsErrCollector));
fprintf('\nmean abs Wout size = %g', mean(WoutAvAbsSizeCollector));
fprintf('\nmax abs err train = %g', max(maxAbsErrTrainCollector));
fprintf('\nmax abs err test = %g', max(maxAbsErrTestCollector));
fprintf('\nNr misclassifieds = %g', sum(misClassNrCollector));
fprintf('\nMax misclassifieds per trial = %g', max(misClassNrCollector));
fprintf('\nNr failed trials = %g\n\n', sum(trialFailCollector));

if plotFlag
    % plot maxErrors
    figure(1); clf;
    set(gcf, 'Position', [900 700 300 250]);
    plot(sort(log2(maxAbsErrTestCollector)), 'LineWidth',2);
    %title('log2 max abs errors test, sorted');
    set(gca, 'FontSize', 12);
    xlabel('trials', 'FontSize', 12);
    ylabel('log2 max abs error', 'FontSize', 12);
    
    % plot exemplary state sequences. If you want to display more than one,
    % repeat execution of the cell below.
    runNr = 0; figNr = 1;
    %%
    runNr = runNr+1;
    thisRunlength = datasetTest(1,3,runNr);
    thisInput = datasetTest(1:thisRunlength, 1:2, runNr);
    stateHarvestPlot = ...
        allstatesNotstartstateLeaky(W, Win, LR, thisInput, startstate);
    figNr = figNr + 1;
    figure(figNr);
    set(gcf, 'Position', [1250 700 300 250]);
    plot(stateHarvestPlot(1:8,:)', 'LineWidth',2);
    xlabel('time (network updates)', 'FontSize', 12);
    ylabel('unit activation', 'FontSize', 12);
    % title(sprintf(['Example state sequence']));
end
