%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% The multiplication task, basic, T = 200 %%%%%%
%%% Here: single trial investigation on %%%%%%%%%%
%%% dynamical mechanism %%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% H. Jaeger, Jan 9, 2014
%
addpath('../share');
set(0,'DefaultFigureWindowStyle','normal');

newModel = 1; newData = 1;

Ntrials = 1; % how many complete train-test trials are to be done.

%%% Set global parameters for a single train-test experiment
T = 200; %  length of trial sequences
Ntrain = 1000;  % Nr of training sequences
Ntest = 100; % Nr of testing sequences
nInternalUnits = 100; % Size of reservoir
connectivity = 0.1; % average connectivity of reservoir 
                     % (1 = fully connected)
inScaling = [.01 1]'; % how the two sets of input weights are scaled
SR = 12; % Weight matrix spectral radius
LR = 0.00005; % smoothing rate, from (0,1]. Low values = much smoothing.
% Value of 1: no smoothing, i.e. reservoir becomes standard
% sigmoid unit network

%%% Create diagnostic collector arrays
genTimeCollector = zeros(1,Ntrials);
learnTimeCollector = zeros(1,Ntrials);
testAvAbsErrCollector = zeros(1,Ntrials);
trainAvAbsErrCollector = zeros(1,Ntrials);
WoutAvAbsSizeCollector = zeros(1,Ntrials);
maxAbsErrTestCollector = zeros(1,Ntrials);
maxAbsErrTrainCollector = zeros(1,Ntrials);
misClassNrCollector = zeros(1,Ntrials);
trialFailCollector = zeros(1,Ntrials);

%%% run experiments
for trialNr = 1:Ntrials
    tAllVar = tic;
    fprintf('%g ', trialNr);
    if mod(trialNr, 20) == 0
        fprintf('\n');
    end
    
    % Generate training and testing data
    if newData
    datasetTrain = genDataMult(T, Ntrain);
    datasetTest = genDataMult(T, Ntest);
    end
    
    % Generate reservoir
    if newModel
        tGenVar = tic;
    WunitSR = generate_internal_weights(nInternalUnits, connectivity);
    tGen = toc(tGenVar);
    WinRaw = 2 * (rand(nInternalUnits, 2) - 0.5);
    end
    
    % Scale weighgts
    W = SR * WunitSR;
    Win = WinRaw * diag(inScaling);
    
    startstate = zeros( nInternalUnits, 1);
    
    % start stopwatch for training
    tTrainVar = tic;
    
    % Collect training state vectors
    stateHarvest = zeros(nInternalUnits, Ntrain);
    teachHarvest = zeros(1, Ntrain);
    for i = 1:Ntrain
        thisRunlength = datasetTrain(1,3,i);
        thisInput = datasetTrain(1:thisRunlength, 1:2, i);
        stateHarvest(:,i) = ...
            laststateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvest(1,i) = datasetTrain(2,3,i);
    end
    
    % Compute output weights
    Wout = (pinv(stateHarvest') * teachHarvest')';
    
    tTrain = toc(tTrainVar);
    
    % Training error diagnostics
    trainOuts = Wout * stateHarvest;
    trainAbsErrs = abs(trainOuts - teachHarvest);
    
    % Compute test outputs
    stateHarvestTest = zeros(nInternalUnits, Ntest);
    teachHarvestTest = zeros(1, Ntest);
    
    for i = 1:Ntest
        thisRunlength = datasetTest(1,3,i);
        thisInput = datasetTest(1:thisRunlength, 1:2, i);
        stateHarvestTest(:,i) = laststateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvestTest(1,i) = datasetTest(2,3,i);
    end
    
    % Performance diagnostics for this trial
    testOuts = Wout * stateHarvestTest;
    avAbsErrTest = mean(abs(teachHarvestTest - testOuts));
    maxAbsErrTest = max(abs(teachHarvestTest - testOuts));
    nrMisclassifieds = sum(abs(teachHarvestTest - testOuts) > 0.04);
    
    
    % Save diagnostics
    genTimeCollector(trialNr) = tGen;
    learnTimeCollector(trialNr) = tTrain;
    trainAvAbsErrCollector(trialNr) = mean(trainAbsErrs);
    maxAbsErrTrainCollector(trialNr) = max(trainAbsErrs);
    testAvAbsErrCollector(trialNr) = avAbsErrTest;
    WoutAvAbsSizeCollector(trialNr) = mean(mean(abs(Wout)));
    maxAbsErrTestCollector(trialNr) = maxAbsErrTest;
    misClassNrCollector(trialNr) = nrMisclassifieds;
    if nrMisclassifieds > Ntest / 100
        trialFailCollector(trialNr) = 1;
    end
    
   
    tAll = toc(tAllVar); % tAll is little below 15 secs here
end


%%% display results
fprintf('\nmean learning time = %g', mean(learnTimeCollector));
fprintf('\nmean av abs error train = %g', mean(trainAvAbsErrCollector));
fprintf('\nmean av abs error test = %g', mean(testAvAbsErrCollector));
fprintf('\nmean abs Wout size = %g', mean(WoutAvAbsSizeCollector));
fprintf('\nmax abs err train = %g', max(maxAbsErrTrainCollector));
fprintf('\nmax abs err test = %g', max(maxAbsErrTestCollector));
fprintf('\nNr misclassifieds = %g', sum(misClassNrCollector));
fprintf('\nMax misclassifieds per trial = %g', max(misClassNrCollector));
fprintf('\nNr failed trials = %g\n\n', sum(trialFailCollector));

% plot maxErrors
figure(1); clf;
set(gcf, 'Position', [900 700 300 250]);
plot(sort(log2(maxAbsErrTestCollector)), 'LineWidth',2);
%title('log2 max abs errors test, sorted');
set(gca, 'FontSize', 12);
xlabel('trials', 'FontSize', 12); 
ylabel('log2 max abs error', 'FontSize', 12);

% plot exemplary state sequences. If you want to display more than one,
% repeat execution of the cell below.
runNr = 0; figNr = 1;
%%
runNr = runNr+1; 
thisRunlength = datasetTest(1,3,runNr);
        thisInput = datasetTest(1:thisRunlength, 1:2, runNr);
        stateHarvestPlot = ...
            allstatesNotstartstateLeaky(W, Win, LR, thisInput, startstate);
figNr = figNr + 1;      
figure(figNr);
set(gcf, 'Position', [1250 700 300 250]);
plot(stateHarvestPlot(1:8,:)', 'LineWidth',2); 
xlabel('time (network updates)', 'FontSize', 12);
ylabel('unit activation', 'FontSize', 12);
% title(sprintf(['Example state sequence']));

%%% Now the dynamics investigations

%% first, check whether Wout is orthogonal to WinRaw(:,1)
disp(sprintf('inner prod Wout WinRaw_1 = %g', ...
    Wout * WinRaw(:,1) / (norm(Wout) * norm(WinRaw(:,1)))));

%% compute (N-1)xN sized projector P on orthogonal subspace to WinRaw
w = WinRaw(:,1); w = w / norm(w);
U = zeros(nInternalUnits,nInternalUnits);
U(:,1) = w;
for k = 2:nInternalUnits
    v = zeros(nInternalUnits,1); v(k) = 1;
    projections = U(:,1:k-1)' * v;
    v = v - U(:,1:k-1) * projections;
    U(:,k) = v / norm(v);    
end
P = U(:,2:end)';

% %% re-execute the test run nr iRun and harvest all states
% iRun = 3;
% thisRunlength = datasetTest(1,3,iRun);
% stateHarvest = zeros(nInternalUnits, thisRunlength);
% x = zeros(nInternalUnits,1);
% for n = 1:thisRunlength
%     thisInput = datasetTest(n, 1:2, iRun);
%     x = (1-LR)*x + LR * tanh(W*x + Win*thisInput');
%     stateHarvest(:,n) = x;
% end
% 
% statesProjected = P * stateHarvest;
% % do SVD
% [U S V] = svd(statesProjected * statesProjected');
% % project on first two PCs for visualization
% statesProjected2D = U(:,1:2)' * statesProjected;
% 
% figure(3);
% set(gcf, 'Position', [900 300 300 250]);
% plot(statesProjected2D(1,:), statesProjected2D(2,:), ...
%     '.', 'LineWidth',2);
% maxVal = max(max(statesProjected2D));
% set(gca, 'XLim', 1.1*[-maxVal maxVal], 'YLim', 1.1* [-maxVal maxVal]);
% xlabel('time (network updates)', 'FontSize', 12);
% ylabel('unit activation', 'FontSize', 12);

% %% carry out  runs with a single cue spike, where the payload
% %% input is ramped
% 
% Inci = 10; % how many increments for ranging u_1 input
% recordTimes = 40:20:200; nRecord = length(recordTimes);
% stateHarvest = zeros(nInternalUnits, Inci, nRecord);
% 
% 
% for u1 = 1:Inci
%     x = zeros(nInternalUnits,1);
%     recCounter = 1;
%     for n = 1:200
%         thisInput = [rand, 0];
%         if n == recordTimes(1)
%             thisInput = [u1/Inci, 1];
%         end
%         x = (1-LR)*x + LR * tanh(W*x + Win*thisInput');
%         if n == recordTimes(recCounter)
%             stateHarvest(:,u1,recCounter) = x;
%             recCounter = recCounter + 1;
%         end
%     end
%     
% end
% 
% % do SVD
% allstatesP = P * reshape(stateHarvest, nInternalUnits, Inci * nRecord);
% [U S V] = svd(allstatesP * allstatesP');
% allstates2D = U(:,2:3)' * allstatesP;
% stateHarvest2D = reshape(allstates2D, 2, Inci, nRecord);
% 
% 
% 
% figure(3); clf;
% set(gcf, 'Position', [900 300 300 250]);
% hold on;
% cm = colormap(jet);
% for i = 1:Inci
%     for t = 1:nRecord
%         plot(stateHarvest2D(1,i,t), stateHarvest2D(2,i,t), ...
%             '.', 'LineWidth',2, 'Color', cm(ceil(i * 64/Inci),:));
%        
%     end
% end
% 
% 
% allstates3D = U(:,1:3)' * allstatesP;
% stateHarvest3D = reshape(allstates3D, 3, Inci, nRecord);
% 
% 
% 
% figure(4); clf;
% set(gcf, 'Position', [1250 300 300 250]);
% hold on;
% cm = colormap(jet);
% for i = 1:Inci
%     for t = 1:nRecord
%         plot3(stateHarvest3D(1,i,t), stateHarvest3D(2,i,t), ...
%             stateHarvest3D(3,i,t), ...
%             '.', 'LineWidth',2, 'Color', cm(ceil(i * 64/Inci),:));
%        
%     end
% end



%% carry out runs with 2 u2 spikes at times 20 and 40, record states after time 40  
Inci = 10; % how many increments for ranging u_1 input
stateHarvest = zeros(nInternalUnits, Inci, Inci);
stateHarvestend = zeros(nInternalUnits, Inci, Inci);


for spike1 = 1:Inci
    for spike2 = 1:Inci
        x = zeros(nInternalUnits,1);
        n1 = randi([1,20]); n2 = randi([21,100]);
        % n1 = 20; n2 = 100;
        for n = 1:200
            thisInput = [rand, 0];
            if n == n1
                thisInput = [spike1/Inci, 1];
            end
            if n == n2
                thisInput = [spike2/Inci, 1];
            end
            x = (1-LR)*x + LR * tanh(W*x + Win*thisInput');
            if n == 100
                stateHarvest(:,spike1,spike2) = x;
            end
            if n == 200
                stateHarvestend(:,spike1,spike2) = x;
            end
        end
    end
end
%%
% do SVD
allstates = reshape(stateHarvest, nInternalUnits, Inci * Inci);
[U S V] = svd(allstates * allstates');
allstates2D = [U(:,[2 ])'; Wout] * allstates;
stateHarvest2D = reshape(allstates2D, 2, Inci, Inci);

allstatesend = reshape(stateHarvestend, nInternalUnits, Inci * Inci);
allstatesend2D = [U(:,[2])'; Wout] * allstatesend;
stateHarvestend2D = reshape(allstatesend2D, 2, Inci, Inci);


figure(3); clf;
set(gcf, 'Position', [900 300 200 300]);
hold on;
cm = colormap(jet);
for s1 = 1:Inci
    for s2 = 1:Inci
    
        plot(stateHarvest2D(1,s1,s2), stateHarvest2D(2,s1,s2), ...
            '.', 'MarkerSize',20, ...
        'Color',cm(ceil(s1 * s2 * 64/Inci^2),:));
   
    end
end
set(gca, 'Box', 'on', 'FontSize', 16, 'YLim', [-0.2 1.2]);




figure(4); clf;
set(gcf, 'Position', [1200 300 200 300]);
hold on;
cm = colormap(jet);
for s1 = 1:Inci
    for s2 = 1:Inci
    
        plot(stateHarvestend2D(1,s1,s2), stateHarvestend2D(2,s1,s2), ...
            '.', 'MarkerSize',20, ...
        'Color',cm(ceil(s1 * s2 * 64/Inci^2),:));
   
    end
end
set(gca, 'Box', 'on', 'FontSize', 16, ...
    'XLim', [-1.5 -1]*10e-5, 'YLim', [-0.2 1.2]);










