%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Rand. permutation task, blind, T 1000, K 150  %%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% H. Jaeger, Feb 10, 2012
%
% This script runs Ntrials many complete training-testing experiments, each
% using fresh data and new random reservoirs.

addpath('../share');
set(0,'DefaultFigureWindowStyle','normal');

saveFlag = 1; plotFlag = 1;

newModel = 1; newData = 1;

Ntrials = 2; % how many complete train-test trials are to be done.

%%% Set global parameters for a single train-test experiment
T = 100; %  length of trial sequences
Ntrain = 500;  % Nr of training sequences
Ntest = 10; % Nr of testing sequences
nInternalUnits = 150; % Size of reservoir
connectivity = 0.066; % average connectivity of reservoir
% (1 = fully connected)
inScaling =  1 * ones(1,100)'; % how the two sets of input weights are scaled
SR = 0.; % Weight matrix spectral radius
LR = 0.000001; % smoothing rate, from (0,1]. Low values = much smoothing.
% Value of 1: no smoothing, i.e. reservoir becomes standard
% sigmoid unit network

%%% Create diagnostic collector arrays
genTimeCollector = zeros(1,Ntrials);
learnTimeCollector = zeros(1,Ntrials);
testAvAbsErrCollector = zeros(1,Ntrials);
trainAvAbsErrCollector = zeros(1,Ntrials);
WoutAvAbsSizeCollector = zeros(1,Ntrials);
maxAbsErrTestCollector = zeros(1,Ntrials);
maxAbsErrTrainCollector = zeros(1,Ntrials);
misClassNrCollector = zeros(1,Ntrials);
trialFailCollector = zeros(1,Ntrials);

%%% run experiments
for trialNr = 1:Ntrials
    tAllVar = tic;
    fprintf('%g ', trialNr);
    if mod(trialNr, 20) == 0
        fprintf('\n');
    end
    
    % Generate training and testing data
    if newData
        datasetTrain = genDataRandPermSparse(T, Ntrain);
        datasetTest = genDataRandPermSparse(2*T, Ntest);
    end
    
    % Generate reservoir
    if newModel
        tGenVar = tic;
        WunitSR = generate_internal_weights(nInternalUnits, connectivity);
        tGen = toc(tGenVar);
        WinRaw = 2 * (rand(nInternalUnits, 100) - 0.5);
    end
    
    % Scale weighgts
    W = SR * WunitSR;
    Win = WinRaw * diag(inScaling);
    
    startstate = zeros( nInternalUnits, 1);
    
    % start stopwatch for training
    tTrainVar = tic;
    
    % Collect training state vectors
    stateHarvest = zeros(nInternalUnits, Ntrain);
    teachHarvest = zeros(100, Ntrain);
    for i = 1:Ntrain
        thisIO = datasetTrain{1,i};
        thisInput = thisIO(:, 1:100);
        stateHarvest(:,i) = ...
            laststateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvest(:,i) = thisIO(end,101:end)';
    end
    
    % Compute output weights
    Wout = (pinv(stateHarvest') * teachHarvest')';
    
    tTrain = toc(tTrainVar);
    
    % Training error diagnostics
    trainOuts = Wout * stateHarvest;
    trainAbsErrs = abs(trainOuts - teachHarvest);
    
    % Compute test outputs
    stateHarvestTest = zeros(nInternalUnits, Ntest);
    teachHarvestTest = zeros(100, Ntest);
    runFailCounter = 0;
    
    for i = 1:Ntest
        thisIO = datasetTest{1,i};
        thisInput = thisIO(:, 1:100);
        stateHarvestTest(:,i) = ...
            laststateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvestTest(:,i) = thisIO(end,101:end)';
        if sum(abs((Wout * stateHarvestTest(:,i)) - ...
                teachHarvestTest(:,i)) > 0.5) > 0
            runFailCounter = runFailCounter + 1;
        end
    end
    
    % Performance diagnostics for this trial
    testOuts = Wout * stateHarvestTest;
    avAbsErrTest = mean(mean(abs(teachHarvestTest - testOuts)));
    maxAbsErrTest = max(max(abs(teachHarvestTest - testOuts)));
    nrMisclassifieds = sum(sum(abs(teachHarvestTest - testOuts) > 0.5));
    nrFailedRuns = runFailCounter;
    
    
    % Save diagnostics
    genTimeCollector(trialNr) = tGen;
    learnTimeCollector(trialNr) = tTrain;
    trainAvAbsErrCollector(trialNr) = mean(mean(trainAbsErrs));
    maxAbsErrTrainCollector(trialNr) = max(max(trainAbsErrs));
    testAvAbsErrCollector(trialNr) = avAbsErrTest;
    WoutAvAbsSizeCollector(trialNr) = mean(mean(abs(Wout)));
    maxAbsErrTestCollector(trialNr) = maxAbsErrTest;
    misClassNrCollector(trialNr) = nrMisclassifieds;
    if nrFailedRuns > Ntest / 100
        trialFailCollector(trialNr) = 1;
    end
    
    
    tAll = toc(tAllVar); % tAll is little above 30 secs here
end

%%% save results
if saveFlag
    save('RandPerm_Blind_T1000K150', 'learnTimeCollector',...
        'genTimeCollector', 'trainAvAbsErrCollector', ...
        'testAvAbsErrCollector', 'WoutAvAbsSizeCollector', ...
        'maxAbsErrTrainCollector', 'maxAbsErrTestCollector',...
        'misClassNrCollector', 'trialFailCollector', 'trialNr');
end

%%% display results
fprintf('\nmean learning time = %g', mean(learnTimeCollector));
fprintf('\nmean av abs error train = %g', mean(trainAvAbsErrCollector));
fprintf('\nmean av abs error test = %g', mean(testAvAbsErrCollector));
fprintf('\nmean abs Wout size = %g', mean(WoutAvAbsSizeCollector));
fprintf('\nmax abs err train = %g', max(maxAbsErrTrainCollector));
fprintf('\nmax abs err test = %g', max(maxAbsErrTestCollector));
fprintf('\nNr misclassifieds = %g', sum(misClassNrCollector));
fprintf('\nMax misclassifieds per trial = %g', max(misClassNrCollector));
fprintf('\nNr failed trials = %g\n\n', sum(trialFailCollector));

