%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Temp order 3 symbol task, basic, T 1000, K 100  %%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% H. Jaeger, Feb 9, 2012
%
% This script runs Ntrials many complete training-testing experiments, each
% using fresh data and new random reservoirs.

addpath('../share');
set(0,'DefaultFigureWindowStyle','normal');

saveFlag = 0; plotFlag = 1;

newModel = 1; newData = 1;

Ntrials = 100; % how many complete train-test trials are to be done.

%%% Set global parameters for a single train-test experiment
T = 1000; %  length of trial sequences
Ntrain = 200;  % Nr of training sequences
Ntest = 1000; % Nr of testing sequences
nInternalUnits = 100; % Size of reservoir
connectivity = 0.1; % average connectivity of reservoir
% (1 = fully connected)
inScaling =  0.001*[1000 1000 1 1 1 1]'; % how the two sets of input weights are scaled
SR = 0.8; % Weight matrix spectral radius
LR = 0.001; % smoothing rate, from (0,1]. Low values = much smoothing.
% Value of 1: no smoothing, i.e. reservoir becomes standard
% sigmoid unit network

%%% Create diagnostic collector arrays
genTimeCollector = zeros(1,Ntrials);
learnTimeCollector = zeros(1,Ntrials);
testAvAbsErrCollector = zeros(1,Ntrials);
trainAvAbsErrCollector = zeros(1,Ntrials);
WoutAvAbsSizeCollector = zeros(1,Ntrials);
maxAbsErrTestCollector = zeros(1,Ntrials);
maxAbsErrTrainCollector = zeros(1,Ntrials);
misClassNrCollector = zeros(1,Ntrials);
trialFailCollector = zeros(1,Ntrials);

%%% run experiments
for trialNr = 1:Ntrials
    tAllVar = tic;
    fprintf('%g ', trialNr);
    if mod(trialNr, 20) == 0
        fprintf('\n');
    end
    
    % Generate training and testing data
    if newData
        datasetTrain = genDataTempOrder3Sym(T, Ntrain);
        datasetTest = genDataTempOrder3Sym(T, Ntest);
    end
    
    % Generate reservoir
    if newModel
        tGenVar = tic;
        WunitSR = generate_internal_weights(nInternalUnits, connectivity);
        tGen = toc(tGenVar);
        WinRaw = 2 * (rand(nInternalUnits, 6) - 0.5);
    end
    
    % Scale weighgts
    W = SR * WunitSR;
    Win = WinRaw * diag(inScaling);
    
    startstate = zeros( nInternalUnits, 1);
    
    % start stopwatch for training
    tTrainVar = tic;
    
    % Collect training state vectors
    stateHarvest = zeros(nInternalUnits, Ntrain);
    teachHarvest = zeros(8, Ntrain);
    for i = 1:Ntrain
        thisInput = datasetTrain(:, 1:6, i);
        stateHarvest(:,i) = ...
            laststateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvest(:,i) = datasetTrain(end,7:14,i)';
    end
    
    % Compute output weights
    Wout = (pinv(stateHarvest') * teachHarvest')';
    
    tTrain = toc(tTrainVar);
    
    % Training error diagnostics
    trainOuts = Wout * stateHarvest;
    trainAbsErrs = abs(trainOuts - teachHarvest);
    
    % Compute test outputs
    stateHarvestTest = zeros(nInternalUnits, Ntest);
    teachHarvestTest = zeros(8, Ntest);
    runFailCounter = 0;
    
    for i = 1:Ntest
        thisInput = datasetTest(:, 1:6,i);
        stateHarvestTest(:,i) = ...
            laststateLeaky(W, Win, LR, thisInput, startstate);
        teachHarvestTest(:,i) = datasetTest(end,7:14,i)';
        if sum(abs((Wout * stateHarvestTest(:,i)) - ...
                teachHarvestTest(:,i)) > 0.5) > 0
            runFailCounter = runFailCounter + 1;
        end
    end
    
    % Performance diagnostics for this trial
    testOuts = Wout * stateHarvestTest;
    avAbsErrTest = mean(mean(abs(teachHarvestTest - testOuts)));
    maxAbsErrTest = max(max(abs(teachHarvestTest - testOuts)));
    nrMisclassifieds = sum(sum(abs(teachHarvestTest - testOuts) > 0.5));
    nrFailedRuns = runFailCounter;
    
    
    % Save diagnostics
    genTimeCollector(trialNr) = tGen;
    learnTimeCollector(trialNr) = tTrain;
    trainAvAbsErrCollector(trialNr) = mean(mean(trainAbsErrs));
    maxAbsErrTrainCollector(trialNr) = max(max(trainAbsErrs));
    testAvAbsErrCollector(trialNr) = avAbsErrTest;
    WoutAvAbsSizeCollector(trialNr) = mean(mean(abs(Wout)));
    maxAbsErrTestCollector(trialNr) = maxAbsErrTest;
    misClassNrCollector(trialNr) = nrMisclassifieds;
    if nrFailedRuns > Ntest / 100
        trialFailCollector(trialNr) = 1;
    end
    
    
    tAll = toc(tAllVar); % tAll is little above 30 secs here
end

%%% save results
if saveFlag
    save('TempOrder3_Basic_T1000K100', 'learnTimeCollector',...
        'genTimeCollector', 'trainAvAbsErrCollector', ...
        'testAvAbsErrCollector', 'WoutAvAbsSizeCollector', ...
        'maxAbsErrTrainCollector', 'maxAbsErrTestCollector',...
        'misClassNrCollector', 'trialFailCollector', 'trialNr');
end

%%% display results
fprintf('\nmean learning time = %g', mean(learnTimeCollector));
fprintf('\nmean av abs error train = %g', mean(trainAvAbsErrCollector));
fprintf('\nmean av abs error test = %g', mean(testAvAbsErrCollector));
fprintf('\nmean abs Wout size = %g', mean(WoutAvAbsSizeCollector));
fprintf('\nmax abs err train = %g', max(maxAbsErrTrainCollector));
fprintf('\nmax abs err test = %g', max(maxAbsErrTestCollector));
fprintf('\nNr misclassifieds = %g', sum(misClassNrCollector));
fprintf('\nMax misclassifieds per trial = %g', max(misClassNrCollector));
fprintf('\nNr failed trials = %g\n\n', sum(trialFailCollector));


% mean learning time = 6.58236
% mean av abs error train = 0.0113893
% mean av abs error test = 0.0237023
% mean abs Wout size = 9.36386e+08
% max abs err train = 0.126613
% max abs err test = 0.344237
% Nr misclassifieds = 0
% Max misclassifieds per trial = 0
% Nr failed trials = 0

% plot maxErrors
if plotFlag
    figure(1); clf;
    set(gcf, 'Position', [900 700 300 250]);
    plot(sort(log2(maxAbsErrTestCollector)), 'LineWidth',2);
    %title('log2 max abs errors test, sorted');
    set(gca, 'FontSize', 12);
    xlabel('trials', 'FontSize', 12);
    ylabel('log2 max abs error', 'FontSize', 12);
    
    % plot exemplary state sequences. If you want to display more than one,
    % repeat execution of the cell below.
    runNr = 0; figNr = 1;
    %%
    runNr = runNr+1;
    thisInput = datasetTest(:, 1:6, runNr);
    stateHarvestPlot = ...
        allstatesNotstartstateLeaky(W, Win, LR, thisInput, startstate);
    figNr = figNr + 1;
    figure(figNr);
    set(gcf, 'Position', [1250 700 300 250]);
    plot(stateHarvestPlot(1:8,:)', 'LineWidth',2);
    xlabel('time (network updates)', 'FontSize', 12);
    ylabel('unit activation', 'FontSize', 12);
    % title(sprintf(['Example state sequence']));
end
